package se.conciliate.documents.impl.test.diskdrive;

import java.io.IOException;
import java.math.BigDecimal;
import java.net.URL;
import java.nio.ByteBuffer;
import java.nio.file.Files;
import java.nio.file.NoSuchFileException;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.attribute.BasicFileAttributes;
import java.nio.file.attribute.UserDefinedFileAttributeView;
import java.time.ZoneId;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.Optional;
import java.util.function.Consumer;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.stream.Collectors;
import java.util.stream.Stream;
import javax.swing.JComponent;
import se.conciliate.extensions.documents.Document;
import se.conciliate.extensions.documents.DocumentFilter;
import se.conciliate.extensions.documents.DocumentFilterType;
import se.conciliate.extensions.documents.DocumentSource;
import se.conciliate.extensions.documents.MetaData;
import se.conciliate.extensions.documents.MetaData.DateTimeValue;
import se.conciliate.extensions.documents.MetaData.ImageValue;
import se.conciliate.extensions.documents.MetaData.NumberValue;
import se.conciliate.extensions.documents.MetaData.TextValue;
import se.conciliate.extensions.documents.MetaData.UrlValue;
import se.conciliate.extensions.documents.MetaDataGroup;
import se.conciliate.extensions.documents.MetaDataType;
import se.conciliate.extensions.store.MTLanguage;
import se.conciliate.extensions.ui.Shell;
import se.conciliate.mt.tools.URLS;

/**
 * This is a simple implementation of the service "DocumentSource". It is registered into
 * OSGi using Declarative Services. This implementation lists files and folders in a root
 * folder that is set by the user through the interface provided from
 * {@link DiskDriveDocumentSource#getEditorComponent(Consumer, String)}.
 * Files typically don't have too much meta data but this source will make use of a few such
 * as modified and created dates, title, file size and the path to the file.
 * This source also showcases a custom meta data called BANANAS, used only to have an editable
 * meta data except for title.
 */
public class DiskDriveDocumentSource implements DocumentSource {

    /**
     * The following are the meta data provided for documents from the DiskDriveDocumentSource.
     * They should be fairly self explanatory. A document source may use any type of meta
     * data as long as they implement {@link MetaData}. This is the meta data for the title
     * of the document.
     */
    private final DiskMetaData.TitleData TITLE = new DiskMetaData.TitleData(this);
    /**
     * The icon meta data used for a document. This meta data and the title meta data are
     * required by all document sources, meaning all documents must have data for these
     * two meta data types.
     */
    private final DiskMetaData.IconData ICON = new DiskMetaData.IconData(this);
    /**
     * Disk documents have a link, or rather, a path on the disk where the file/document
     * resides. This meta data describes the link.
     */
    private final DiskMetaData.LinkData LINK = new DiskMetaData.LinkData(this);
    /**
     * Disk documents have a type, such as FOLDER and FILE. Other document sources may
     * have types and if they do they can be things like Template or Instruction and so on.
     */
    private final DiskMetaData.TypeData TYPE = new DiskMetaData.TypeData(this);
    /**
     * Meta data describing when the file was created.
     */
    private final DiskMetaData.CreatedData CREATED = new DiskMetaData.CreatedData(this);
    /**
     * Meta data describing when the file was modified.
     */
    private final DiskMetaData.ModifiedData MODIFIED = new DiskMetaData.ModifiedData(this);
    /**
     * Meta data describing the size of the file.
     */
    private final DiskMetaData.FileSizeData FILESIZE = new DiskMetaData.FileSizeData(this);
    /**
     * Meta data showcasing an editable meta data field that the user can edit from Modeling Tool.
     * It is just a number set on a file (if the type of the file is an image). Not tested on
     * other platforms than Windows 10 (different file systems may not have the ability for
     * applications to set its own data on files).
     */
    private final DiskMetaData.BananaData BANANAS = new DiskMetaData.BananaData(this);

    /**
     * The unique id of this source. Used to separate documents from different sources.
     */
    private final String ID = "se.conciliate.DiskDriveDocumentStoreTest";

    /**
     * The current settings. Contains the paths to the roots where files and folders can be
     * found.
     */
    private DiskDriveSettings settings;
    /**
     * The shell is a service from Modeling Tool. The shell represents the state of the UI such
     * as the current main window (mainFrame). It is mostly used in this example code to get a
     * handle to the main frame and use that main frame to position the file chooser UI.
     */
    private Shell shell;

    /**
     * Services provided by Modeling Tool are injected. What to inject is configured in this
     * project's file /resources/OSGI-INF/service-declaration.xml.
     * @param shell
     */
    public void setShell(Shell shell) {
        this.shell = shell;
    }

    @Override
    public String getID() {
        return ID;
    }

    @Override
    public String getTitle() {
        return "DiskDrive Documents";
    }

    @Override
    public URL getIcon() {
        // Modeling Tool registers its own URL-handler for the icon-scheme. The
        // icon can be any icon though, including icons found on http-urls. Only
        // png and jpg icons are currently supported.
        return URLS.createURL("icon:o24/hard_drive.png");
    }

    @Override
    public List<DocumentFilterType> getFilterTypes() {
        // Filters are used by Modeling Tool's browser. Disk drive documents only
        // have one filter that filters on the "file"-type. A file is either a
        // directory or a file and this filter is used to filter on that information,
        // so the browser can be made to show either everything in the roots, just the
        // directories or just the files.
        return Arrays.asList(new FileTypeFilter(this));
    }

    @Override
    public List<MetaDataGroup> getMetaDataGroups(Document doc) {
        // A meta data group is mainly used to structure the UI of the document editor.
        // The document editor is where Modeling Tool displays all data for a document
        // and if the meta data is editable, where that data is edited. To see more,
        // look at the interface MetaDataGroup. Basically, each group will get its own
        // side tab in the editor. The disk meta data only creates one group called general.
        // In this tab, all meta data is placed.
        // You may wonder where the Title and Icon meta data are. Since these are required, they
        // are always displayed in the editor and will not be shown in a tab. Open the document
        // editor to see where they end up. (To open the editor, first configure the settings
        // in modeling tool, then open the browser and make sure it is showing document sources
        // "Dokumentkällor" in swe. Then double click on a file/directory.)
        return Arrays.asList(
                MetaDataGroup.createGroup("General", URLS.createURL("icon:t32/document_empty.png"))
                    .addMetaDataType(LINK)
                    .addMetaDataType(TYPE)
                    .addMetaDataType(CREATED)
                    .addMetaDataType(MODIFIED)
                    .addMetaDataType(FILESIZE)
                    .addMetaDataType(BANANAS)
        );
    }

    @Override
    public List<MetaDataType<? extends MetaData<?>>> getMetaDataTypes() {
        // This should return ALL meta data types that this plugin need to register.
        return new ArrayList<>(Arrays.<MetaDataType<? extends MetaData<?>>>asList(TITLE, ICON, LINK, TYPE, CREATED, MODIFIED, FILESIZE, BANANAS));
    }

    @Override
    public List<String> getDocumentIDs(Collection<DocumentFilter> filters, MetaDataType orderBy, boolean ascendingOrder) {
        List<Path> files = new ArrayList<>();
        for(DiskDriveSettings.DiskDrivePath path : settings.getPaths()) {
            files.addAll(listContent(Paths.get(path.getPath()), path.isRecursive(), path.isListDirectories()));
        }
        
        for(DocumentFilter filter : filters) {
            if(filter.isFiltering() && filter instanceof DiskFilter) {
                DiskFilter f = (DiskFilter)filter;
                files = files.stream().filter(f::isMatch).collect(Collectors.toList());
            }
        }
        
        return files.stream().map(path -> path.toAbsolutePath())
                .sorted((a, b) -> {
                    int comp = a.getFileName().compareTo(b.getFileName());
                    return comp == 0 ? a.compareTo(b) : comp;
                })
                .map(Path::toString)
                .collect(Collectors.toList());
    }

    @Override
    public List<Document> getDocuments(List<String> ids, Collection<? extends MetaDataType<?>> columns) {
        return ids.stream()
            .map(path -> Paths.get(path))
            .filter(Files::exists)
            .map(path -> Document.createExisting(this, path.toAbsolutePath().toString(), meta(path)))
            .collect(Collectors.toList());
    }

    @Override
    public JComponent getEditorComponent(Consumer<String> callback, String current) {
        // Returns the settings editor.
        return new DiskDriveSettingsEditor(shell.getShellFrame(), callback, new DiskDriveSettings(current));
    }

    @Override
    public void initialize(String settingsOrEmpty) {
        // Called from Modeling Tool once the framework is ready to start using this source.
        update(settingsOrEmpty);
    }

    @Override
    public void update(String settings) {
        this.settings = new DiskDriveSettings(settings);
    }

    @Override
    public Optional<Creator> getCreator() {
        // Used to create new documents from this source. DiskDriveDocuments cannot be created,
        // even though it would be fairly simple to implement, so we just return empty (no creator).
        return Optional.empty();
    }

    @Override
    public Optional<Remover> getRemover(Collection<Document> documents) {
        // Used to remove documents from this source. DiskDriveDocuments cannot be removed,
        // even though it would be fairly simple to implement, so we just return empty (no remover).
        return Optional.empty();
    }

    @Override
    public Optional<Saver> getSaver() {
        // Used to save documents from this source. Since we have an editable meta data in the form
        // of BANANAS, we need to be able to save changes to this data. Since Saver is a functional
        // interface we just return an optional (Java's monad-ish Maybe-like implementation) of
        // a function that can save bananas on a file.
        return Optional.of(documents -> {
            documents.forEach(doc -> {
                Number bananas = doc.getMetaData(BANANAS).map(data -> data.getData()).orElse(null);
                Path path = doc.getMetaData(LINK).map(data -> data.getData()).map(link -> Paths.get(link)).orElse(null);
                if(bananas != null && path != null) {
                    write(bananas.toString(), path);
                }
            });
            return true;
        });
    }

    @Override
    public Optional<DocumentFilter> getPatternFilter(String pattern, MTLanguage language, boolean exactMatch, boolean ignoreCase) {
        return Optional.of(new DiskPatternFilter(pattern, language, exactMatch, ignoreCase));
    }

    private Collection<MetaData> meta(Path path) {
        Collection<MetaData> data = new ArrayList<>();
        data.add(new TextValue(TITLE, path.getFileName().toString()));
        data.add(new ImageValue(ICON, URLS.createURL("icon:o24/hard_drive.png")));
        data.add(new UrlValue(LINK, path.toUri().toString()));
        data.add(new MetaData.OptionsValue(TYPE, Arrays.asList(Files.isDirectory(path) ? "directory" : "file")));
        data.add(new NumberValue(FILESIZE, path.toFile().length() / 1024));

        String bananas = read(path);
        if(bananas != null) {
            data.add(new NumberValue(BANANAS, new BigDecimal(bananas)));
        }

        try {
            BasicFileAttributes attr = Files.readAttributes(path, BasicFileAttributes.class);
            data.add(new DateTimeValue(CREATED, attr.creationTime().toInstant()
                    .atZone(ZoneId.systemDefault()).toLocalDateTime()));
            data.add(new DateTimeValue(MODIFIED, attr.lastModifiedTime().toInstant()
                    .atZone(ZoneId.systemDefault()).toLocalDateTime()));
        } catch(IOException ex) {
            Logger.getLogger(getClass().getName()).log(Level.WARNING, null, ex);
        }

        return data;
    }
    
    private List<Path> listContent(Path root, boolean recursive, boolean listDirectories) {
        List<Path> paths = new ArrayList<>();
        
        try(Stream<Path> fileStream = recursive ? 
            Files.walk(root)
                .filter(path -> !path.equals(root))
                .filter(path -> listDirectories ? true : Files.isRegularFile(path)) :
            Files.list(root)
                .filter(path -> listDirectories ? true : Files.isRegularFile(path))) {
            
            fileStream.forEach(paths::add);
                
        } catch(IOException ex) {
            Logger.getLogger(getClass().getName()).log(Level.WARNING, "Failed to list files.", ex);
        }
        
        return paths;
    }

    private void write(String value, Path p) {
        UserDefinedFileAttributeView view = Files.getFileAttributeView(p, UserDefinedFileAttributeView.class);
        if(view != null) {
            try {
                byte[] bytes = value.getBytes("UTF-8");
                ByteBuffer bb = ByteBuffer.allocate(bytes.length);
                bb.put(bytes);
                view.write("se.conciliate.mt.banana", bb.flip());
            } catch(IOException ex) {
                Logger.getLogger(getClass().getName()).log(Level.SEVERE, "", ex);
            }
        }
    }

    private String read(Path p) {
        UserDefinedFileAttributeView view = Files.getFileAttributeView(p, UserDefinedFileAttributeView.class);
        if(view != null) {
            try {
                ByteBuffer bb = ByteBuffer.allocate(view.size("se.conciliate.mt.banana"));
                view.read("se.conciliate.mt.banana", bb);
                return new String(bb.flip().array(), "UTF-8");
            } catch(IOException ex) {
                if(!(ex instanceof NoSuchFileException && ex.getMessage().contains("se.conciliate.mt.banana"))) {
                    Logger.getLogger(getClass().getName()).log(Level.SEVERE, "", ex);
                }
            }
        }
        return null;
    }
}
