package se.conciliate.documents.impl.test.diskdrive;

import java.io.StringReader;
import java.util.ArrayList;
import java.util.List;
import java.util.stream.Collectors;
import javax.json.Json;
import javax.json.JsonArrayBuilder;
import javax.json.JsonObject;
import javax.json.JsonReader;
import javax.json.JsonStructure;
import javax.json.JsonValue.ValueType;

/**
 * The settings class for disk drive document sources. A simple Json-string listing
 * all paths on a file system that should act as root directories where files and
 * directories can be found.
 */
public class DiskDriveSettings {

    private final static String DEFAULT_SETTINGS = "{\"paths\": []}";

    public static final class DiskDrivePath {
        private final String path;
        private final boolean recursive;
        private final boolean listDirectories;

        public DiskDrivePath(String path, boolean recursive, boolean listDirectories) {
            this.path = path;
            this.recursive = recursive;
            this.listDirectories = listDirectories;
        }

        public String getPath() {
            return path;
        }

        public boolean isRecursive() {
            return recursive;
        }

        public boolean isListDirectories() {
            return listDirectories;
        }
        
    }
    
    private final List<DiskDrivePath> paths;

    public DiskDriveSettings(List<DiskDrivePath> paths) {
        this.paths = paths;
    }
    
    public DiskDriveSettings(String settings) {
        settings = (settings == null ? DEFAULT_SETTINGS : (settings.isEmpty() ? DEFAULT_SETTINGS : settings));
        JsonReader reader = Json.createReader(new StringReader(settings));
        JsonStructure structure = reader.read();
        switch(structure.getValueType()) {
            case OBJECT:
                this.paths = ((JsonObject)structure)
                    .getJsonArray("paths")
                    .stream()
                    .filter(path -> path.getValueType() == ValueType.OBJECT)
                    .map(JsonObject.class::cast)
                    .map(path -> new DiskDrivePath(
                            path.getString("path"),
                            path.getBoolean("recursive"),
                            path.getBoolean("list-directories")
                    ))
                    .collect(Collectors.toList());
                    break;
                default:
                    this.paths = new ArrayList<>();
                    break;
        }
    }
    
    public List<DiskDrivePath> getPaths() {
        return paths;
    }
    
    public String toJson() {
        JsonArrayBuilder list = Json.createArrayBuilder();
        getPaths().forEach(path -> {
            JsonObject pathData = Json.createObjectBuilder()
                .add("path", path.getPath())
                .add("recursive", path.isRecursive())
                .add("list-directories", path.isListDirectories())
                .build();
            list.add(pathData);
        });
        return Json.createObjectBuilder()
                .add("paths", list)
                .build()
                .toString();
    }
    
}
