package se.conciliate.documents.impl.test.diskdrive;

import java.math.BigDecimal;
import java.time.format.DateTimeFormatter;
import java.util.Optional;
import se.conciliate.extensions.documents.Document;
import se.conciliate.extensions.documents.DocumentSource;
import se.conciliate.extensions.documents.InputDescriptor;
import se.conciliate.extensions.documents.InputDescriptor.DateTimeDescriptor;
import se.conciliate.extensions.documents.InputDescriptor.NumberDescriptor;
import se.conciliate.extensions.documents.InputDescriptor.OptionsDescriptor;
import se.conciliate.extensions.documents.MetaData;
import se.conciliate.extensions.documents.MetaData.DateTimeValue;
import se.conciliate.extensions.documents.MetaData.ImageValue;
import se.conciliate.extensions.documents.MetaData.NumberValue;
import se.conciliate.extensions.documents.MetaData.OptionsValue;
import se.conciliate.extensions.documents.MetaData.TextValue;
import se.conciliate.extensions.documents.MetaData.UrlValue;
import se.conciliate.extensions.documents.MetaDataType;
import se.conciliate.mt.tools.URLS;
import se.conciliate.mt.ui.laf.HiDpiIcon;

public abstract class DiskMetaData<T extends MetaData<?>> implements MetaDataType<T> {

    private final DiskDriveDocumentSource source;
    private final String title;

    private DiskMetaData(DiskDriveDocumentSource source, String title) {
        this.source = source;
        this.title = title;
    }

    @Override
    public DocumentSource getDocumentSource() {
        return source;
    }

    @Override
    public String getTitle() {
        return title;
    }

    public static class TitleData extends DiskMetaData<TextValue> implements TextMetaDataType {
        public TitleData(DiskDriveDocumentSource source) {
            super(source, "Title");
        }

        @Override
        public String getID() {
            return "disk.title";
        }
        
        @Override
        public Class<TextValue> getDataType() {
            return TextValue.class;
        }

        @Override
        public ColumnType getColumnType() {
            return ColumnType.TITLE;
        }
    }
    
    public static class TypeData extends DiskMetaData<OptionsValue> implements OptionsMetaDataType {
        public TypeData(DiskDriveDocumentSource source) {
            super(source, "Type");
        }

        @Override
        public String getID() {
            return "disk.type";
        }
        
        @Override
        public Class<OptionsValue> getDataType() {
            return OptionsValue.class;
        }

        @Override
        public ColumnType getColumnType() {
            return ColumnType.DOCUMENT_TYPES;
        }

        @Override
        public Optional<OptionsDescriptor> getInputDescriptor() {
            return Optional.of(InputDescriptor.options(this)
                    .singleSelection()
                    .option("file", "File", new HiDpiIcon(URLS.createURL("icon:t24/document_empty.png")))
                    .option("directory", "Directory", new HiDpiIcon(URLS.createURL("icon:t24/folder.png")))
                    .build());
        }
        
    }

    public static class IconData extends DiskMetaData<ImageValue> implements ImageMetaDataType {
        public IconData(DiskDriveDocumentSource source) {
            super(source, "Icon");
        }
        
        @Override
        public String getID() {
            return "disk.icon";
        }

        @Override
        public Class<ImageValue> getDataType() {
            return ImageValue.class;
        }

        @Override
        public ColumnType getColumnType() {
            return ColumnType.TITLE;
        }
    }

    public static class FileSizeData extends DiskMetaData<NumberValue> implements NumberMetaDataType {

        FileSizeData(DiskDriveDocumentSource source) {
            super(source, "File size");
        }

        @Override
        public String getID() {
            return "disk.filesize";
        }

        @Override
        public Class<NumberValue> getDataType() {
            return NumberValue.class;
        }

        @Override
        public Optional<NumberDescriptor> getInputDescriptor() {
            return Optional.of(InputDescriptor.number(this)
                .decimalPoint('^')
                .unit("KB")
                .build());
        }
    }

    public static class BananaData extends DiskMetaData<NumberValue> implements NumberMetaDataType {

        BananaData(DiskDriveDocumentSource source) {
            super(source, "Bananas");
        }

        @Override
        public String getID() {
            return "disk.banana";
        }

        @Override
        public Class<NumberValue> getDataType() {
            return NumberValue.class;
        }

        @Override
        public Optional<NumberDescriptor> getInputDescriptor() {
            return Optional.of(InputDescriptor.number(this)
                    .decimalPoint('~')
                    .unit("bananas")
                    .decimalCount(3)
                    .maxValue(1000)
                    .minValue(-100)
                    .build());
        }

        @Override
        public boolean isEditable(Document doc) {
            String path = (String)doc.getMetaData("disk.path").map(data -> data.getData()).orElse("");
            if(path.endsWith(".jpg")) {
                return true;
            }
            return false;
        }

        @Override
        public NumberValue getDefaultValue(Document doc) {
            String path = (String)doc.getMetaData("disk.path").map(data -> data.getData()).orElse("");
            if(path.endsWith(".jpg")) {
                return new NumberValue(this, new BigDecimal(0));
            }
            throw new IllegalArgumentException("That document does not have bananas.");
        }


    }

    public static class LinkData extends DiskMetaData<UrlValue> implements UrlMetaDataType {
        public LinkData(DiskDriveDocumentSource source) {
            super(source, "Path");
        }
        
        @Override
        public String getID() {
            return "disk.path";
        }

        @Override
        public Class<UrlValue> getDataType() {
            return UrlValue.class;
        }

        @Override
        public ColumnType getColumnType() {
            return ColumnType.LINK;
        }
    }

    public static class CreatedData extends DiskMetaData<DateTimeValue> implements DateTimeMetaDataType {
        public CreatedData(DiskDriveDocumentSource source) {
            super(source, "Created");
        }
        
        @Override
        public String getID() {
            return "disk.created";
        }

        @Override
        public Class<DateTimeValue> getDataType() {
            return DateTimeValue.class;
        }

        @Override
        public Optional<DateTimeDescriptor> getInputDescriptor() {
            return Optional.of(InputDescriptor.dateTime(this)
                    .formatter(DateTimeFormatter.ofPattern("yyyy-MM-dd hh:mm"))
                    .build());
        }

        @Override
        public ColumnType getColumnType() {
            return ColumnType.CREATED;
        }
        
    }
    
    public static class ModifiedData extends DiskMetaData<DateTimeValue> implements DateTimeMetaDataType {
        public ModifiedData(DiskDriveDocumentSource source) {
            super(source, "Last modified");
        }
        
        @Override
        public String getID() {
            return "disk.modified";
        }

        @Override
        public Class<DateTimeValue> getDataType() {
            return DateTimeValue.class;
        }

        @Override
        public Optional<DateTimeDescriptor> getInputDescriptor() {
            return Optional.of(InputDescriptor.dateTime(this)
                    .formatter(DateTimeFormatter.ofPattern("yyyy-MM-dd hh:mm"))
                    .build());
        }

        @Override
        public ColumnType getColumnType() {
            return ColumnType.LAST_MODIFIED;
        }
        
    }
}
