package se.conciliate.documents.impl.test.diskdrive;

import java.io.File;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;
import javax.swing.Icon;
import se.conciliate.extensions.documents.DocumentFilter;
import se.conciliate.extensions.documents.DocumentFilterType;
import se.conciliate.extensions.documents.DocumentSource;
import se.conciliate.extensions.store.query.FilterEditorType;
import se.conciliate.mt.tools.URLS;
import se.conciliate.mt.ui.laf.HiDpiIcon;

public class FileTypeFilter implements DocumentFilterType {
    private final DiskDriveDocumentSource source;

    public FileTypeFilter(DiskDriveDocumentSource source) {
        this.source = source;
    }
    
    @Override
    public String getID() {
        return "se.conciliate.diskdrive.typefilter";
    }

    @Override
    public DocumentSource getDocumentSource() {
        return source;
    }
    
    @Override
    public String getTitle() {
        return "File type";
    }

    @Override
    public Optional<Icon> getIcon() {
        return Optional.of(new HiDpiIcon("icon:t24/folder_document2.png"));
    }

    @Override
    public DocumentFilter createFilter() {
        return new Filter(null);
    }

    @Override
    public DocumentFilter createFilter(String filterString) {
        return new Filter(filterString);
    }

    private class Filter implements DocumentFilter, DiskFilter {
        private final List<String> available = Arrays.asList("File", "Directory");
        private List<Object> selected = new ArrayList<>();

        public Filter(String filter) {
            if(filter != null && !filter.isEmpty()) {
                Arrays.asList(filter.split(":")).stream()
                        .forEach(selected::add);
            }
        }

        @Override
        public boolean isMatch(Path path) {
            if(Files.isDirectory(path)) {
                return selected.contains("Directory");
            } else if(Files.isRegularFile(path)) {
                return selected.contains("File");
            } else {
                return false;
            }
        }
        
        @Override
        public DocumentFilterType getFilterType() {
            return FileTypeFilter.this;
        }

        @Override
        public String getFilterString() {
            return selected.stream()
                    .map(Object::toString)
                    .collect(Collectors.joining(":"));
        }

        @Override
        public String getFilterRepresentation() {
            String s = FileTypeFilter.this.getTitle() + ": ";
            if(getSelectedKeys().isEmpty()) {
                s += "All";
            } else {
                s += getSelectedKeys().stream().map(this::getTitle).collect(Collectors.joining(", "));
            }
            return s;
        }

        @Override
        public boolean isFiltering() {
            return !selected.isEmpty();
        }

        @Override
        public FilterEditorType getEditorType() {
            return FilterEditorType.SELECTED_KEYS;
        }

        @Override
        public List<Object> getAvailableKeys() {
            return new ArrayList<>(available);
        }

        @Override
        public String getTitle(Object key) {
            return key.toString();
        }

        @Override
        public Icon getIcon(Object key) {
            if("File".equals(key)) {
                return new HiDpiIcon(URLS.createURL("icon:t24/document_header.png"));
            } else {
                return new HiDpiIcon(URLS.createURL("icon:t24/folder.png"));
            }
        }

        @Override
        public List<Object> getSelectedKeys() {
            return Collections.unmodifiableList(selected);
        }

        @Override
        public void setSelectedKeys(List<Object> selectedKeys) {
            this.selected = new ArrayList<>(selectedKeys);
        }
        
    }
    
}
